from pathlib import Path
import subprocess as sp
import sys
from typing import Any, Callable, Iterable, Self, Sequence
import os

import inquirer


cache = Path("~/.cache/kisiac").expanduser()


def is_in_tmux_or_screen() -> bool:
    term = os.environ.get("TERM", "unknown")
    return term.startswith("tmux") or term.startswith("screen")


def as_list(method: Callable[..., Iterable]) -> Callable[..., list]:
    def wrapper(*args: Any, **kwargs: Any) -> list:
        return list(method(*args, **kwargs))

    return wrapper


def multiline_input(msg: str) -> str:
    lines = []
    print(msg, "Hit Ctrl+D to finish.")
    while True:
        try:
            lines.append(input())
        except EOFError:
            return "\n".join(lines)


def handle_key_error(msg: str) -> Callable:
    def decoator(func: Callable) -> Callable:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except KeyError as e:
                raise UserError(f"{msg}: {e}") from e

        return wrapper

    return decoator


class Singleton:
    @classmethod
    def get_instance(cls, *args, **kwargs) -> Self:
        if not hasattr(cls, "_instance") or cls._instance is None:
            cls._instance = cls(*args, **kwargs)
        return cls._instance


def confirm_action(desc: str) -> bool:
    from kisiac.runtime_settings import GlobalSettings

    if GlobalSettings.get_instance().non_interactive:
        return True

    log_msg(desc)
    response = inquirer.prompt(
        [
            inquirer.List(
                "action", message="Proceed?", choices=["yes", "no"], default="no"
            )
        ]
    )
    assert response is not None
    return response["action"] == "yes"


def provide_password(msg: str) -> str:
    from kisiac.runtime_settings import GlobalSettings

    password = os.environ.get("KISIAC_ENCRYPTION_PASSWORD")

    if GlobalSettings.get_instance().non_interactive and password is None:
        raise UserError(
            "No password provided via KISIAC_ENCRYPTION_PASSWORD but non-interactive mode activated."
        )

    if password is not None:
        return password

    while True:
        response = inquirer.prompt(
            [
                inquirer.Password("password", message=msg),
                inquirer.Password("confirm_password", message="Confirm password"),
            ]
        )
        assert response is not None
        if response["password"] != response["confirm_password"]:
            log_msg("Passwords do not match, try again.")
        else:
            return response["password"]


def exists_cmd(cmd: str, host: str, sudo: bool) -> bool:
    try:
        run_cmd(["which", cmd], host=host, sudo=sudo, user_error=False)
        return True
    except sp.CalledProcessError:
        return False


def log_msg(*msgs: Any, host: str | None = None) -> None:
    ann_msgs = list(msgs)
    if host is not None:
        ann_msgs = [f"[{host}]", *msgs]
    print(" ".join(map(str, ann_msgs)), file=sys.stderr)


def cmd_to_str(*cmds: Sequence[str]) -> str:
    return "\n".join(" ".join(map(str, cmd)) for cmd in cmds)


def run_cmd(
    cmd: Sequence[str | Path],
    input: str | None = None,
    host: str = "localhost",
    env: dict[str, Any] | None = None,
    sudo: bool = False,
    user_error: bool = True,
    user_error_msg: str = "",
    check: bool = True,
) -> sp.CompletedProcess[str]:
    """Run a system command using subprocess.run and check for errors."""

    def fmt_cmd_item(item: str | Path) -> str:
        str_item = str(item)
        if " " in str_item or "\t" in str_item:
            if '"' in str_item:
                raise UserError(
                    f'command item {str_item} contains "-characters. This is currently not supported.'
                )
            str_item = f'"{str_item}"'
        return str_item

    postprocesed_cmd: Sequence[str] = list(map(fmt_cmd_item, cmd))
    if sudo:
        postprocesed_cmd = ["sudo", "bash", "-c", f"{' '.join(postprocesed_cmd)}"]
    if host != "localhost":
        if sudo:
            postprocesed_cmd = [
                "ssh",
                host,
                f"sudo bash -c '{' '.join(postprocesed_cmd)}'",
            ]
        else:
            postprocesed_cmd = ["ssh", host, f"{' '.join(postprocesed_cmd)}"]
    log_msg("Running command", cmd_to_str(postprocesed_cmd), host=host)
    try:
        return sp.run(
            postprocesed_cmd,
            check=check,
            text=True,
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            input=input,
            env=env,
        )
    except sp.CalledProcessError as e:
        if user_error:
            if user_error_msg:
                user_error_msg += ": "
            err = e.stderr
            if not err:
                err = e.stdout
            raise UserError(
                f"{user_error_msg}Error occurred while running command '{' '.join(postprocesed_cmd)}': {err}"
            ) from e
        else:
            raise


class UserError(Exception):
    """Base class for user-related errors."""

    pass


def check_type(item: str, value: Any, expected_type: Any) -> None:
    if not isinstance(value, expected_type):
        raise UserError(
            f"Expecting {expected_type} for {item}, found "
            f"value {value} of type {type(value)}."
        )


class HostAgnosticPath:
    def __init__(
        self, path: str | Path, host: str = "localhost", sudo: bool = False
    ) -> None:
        self.path = Path(path)
        self.host = host
        self.sudo = sudo
        if self.is_local_and_user():
            # if non local or sudo, shell commands will be used, which expand
            # the ~ operator automatically
            self.path = self.path.expanduser()

    def read_text(self) -> str:
        if self.is_local_and_user():
            return self.path.read_text()
        else:
            return self._run_cmd(["cat", str(self.path)]).stdout

    def write_text(self, content: str) -> None:
        if self.is_local_and_user():
            self.path.write_text(content)
        else:
            self._run_cmd(
                ["tee", str(self.path)],
                input=content,
            )

    def mkdir(self) -> None:
        if self.is_local_and_user():
            self.path.mkdir(parents=True, exist_ok=True)
        else:
            self._run_cmd(["mkdir", "-p", str(self.path)])

    def chmod(self, *mode: str, recursive: bool = True) -> None:
        self._chperm("chmod", ",".join(mode), recursive=recursive)

    def chown(self, user: str | None, group: str | None = None) -> None:
        if user is not None:
            owner = f"{user}:{group}" if group else user
            self._chperm("chown", owner)
        elif group is not None:
            self._chperm("chgrp", group)
        else:
            raise ValueError("Either user or group must be provided.")

    def setfacl(self, *mode: str, recursive: bool = True, default: bool = True) -> None:
        self._chperm(
            "setfacl",
            *(["-d"] if default else []),
            "-m",
            ",".join(mode),
            recursive=recursive,
        )

    def _chperm(self, cmd: str, *args: str, recursive: bool = True) -> None:
        if recursive and self.is_dir():
            args = ("-R", *args)
        self._run_cmd([cmd, *args, str(self.path)])

    def is_local_and_user(self) -> bool:
        return self.host == "localhost" and not self.sudo

    def _run_cmd(
        self, cmd: Sequence[str], input: str | None = None, user_error: bool = True
    ) -> sp.CompletedProcess[str]:
        return run_cmd(
            cmd,
            input=input,
            host=self.host,
            sudo=self.sudo,
            user_error=user_error,
        )

    def exists(self) -> bool:
        if self.is_local_and_user():
            return self.path.exists()
        else:
            try:
                self._run_cmd(
                    ["test", "-e", str(self.path)],
                    user_error=False,
                )
                return True
            except sp.CalledProcessError:
                return False

    def is_dir(self) -> bool:
        if self.is_local_and_user():
            return self.path.is_dir()
        else:
            try:
                self._run_cmd(
                    ["test", "-d", str(self.path)],
                    user_error=False,
                )
                return True
            except sp.CalledProcessError:
                return False

    def with_suffix(self, suffix: str) -> Self:
        return type(self)(self.path.with_suffix(suffix), host=self.host, sudo=self.sudo)

    @property
    def parents(self) -> Sequence[Self]:
        return [
            type(self)(parent, host=self.host, sudo=self.sudo)
            for parent in self.path.parents
        ]

    def __truediv__(self, other: Any) -> Self:
        return type(self)(self.path / other, host=self.host, sudo=self.sudo)

    def __rtruediv__(self, other: Any) -> Self:
        return type(self)(other / self.path, host=self.host, sudo=self.sudo)

    def __str__(self) -> str:
        if self.host == "localhost":
            return str(self.path)
        else:
            return f"{self.host}:{self.path}"



from collections import defaultdict
from dataclasses import dataclass, field
from enum import StrEnum
import json
from pathlib import Path
import re
from typing import Any, Iterable, Self

from humanfriendly import parse_size

from kisiac.common import check_type, exists_cmd, run_cmd, UserError


CRYPT_PREFIX = "crypt_"
VGS_DEVICE_REPORT_RE = re.compile(r"^(?P<device>.+)\((?P<info>.+)\)$")
LV_INTERNAL_NAME = re.compile(r"^\[(?P<name>.+)\]$")


def parse_lv_name(name_entry) -> tuple[str, bool]:
    m = LV_INTERNAL_NAME.match(name_entry)
    if m:
        return m.group("name"), True
    else:
        return name_entry, False


@dataclass(frozen=True)
class PV:
    device: Path


class Subvolume(StrEnum):
    ORIG = "corig"
    CACHE = "cvol"


def _subvolume_name(lv_name: str, subvolume: Subvolume) -> str:
    return f"{lv_name}_{subvolume}"


@dataclass
class LV:
    name: str
    layout: set[str]
    size: int | None
    stripes: int
    stripe_size: int  # in bytes
    pv_tag: str | None
    cache_size: int | None
    cache_pv_tag: str | None
    cache_mode: str | None = None

    @property
    def cache_subvolume_name(self) -> str:
        return _subvolume_name(self.name, Subvolume.CACHE)

    @property
    def orig_subvolume_name(self) -> str:
        return _subvolume_name(self.name, Subvolume.ORIG)

    def is_cached(self) -> bool:
        return self.cache_mode is not None

    def is_same_layout(self, other: Self) -> bool:
        # TODO find a better way to compare the layout in the cache case
        if "cache" in other.layout:
            # current LV is cached, original layout is not reported
            return True
        return (
            self.layout <= other.layout
            or other.layout <= self.layout
            and self.stripes == other.stripes
            and self.stripe_size == other.stripe_size
        )

    def is_same_size(self, other: Self) -> bool:
        if self.fills_vg() or other.fills_vg():
            # if both just fill the VG, their sizes are considered unchanged
            return self.fills_vg() and other.fills_vg()

        assert self.size is not None and other.size is not None

        def simplify(size: int) -> int:
            # tens of MB should be precise enough
            return size // 10**7

        return simplify(self.size) == simplify(other.size)

    def fills_vg(self) -> bool:
        return self.size is None

    def size_arg(self) -> list[str]:
        return self._size_arg(Subvolume.ORIG)

    def cache_size_arg(self) -> list[str]:
        return self._size_arg(Subvolume.CACHE)

    def _size_arg(self, subvolume: Subvolume) -> list[str]:
        size = self.size if subvolume is Subvolume.ORIG else self.cache_size
        if size is None:
            assert subvolume is Subvolume.ORIG, "Cache resizing not supported"
            return ["--extents", "+100%FREE"]
        else:
            arg = "--cachesize" if subvolume is Subvolume.CACHE else "--size"
            return [arg, f"{size}B"]

    def stripe_args(self) -> list[str]:
        if self.stripes > 1:
            return [
                "--stripes",
                str(self.stripes),
                "--stripesize",
                f"{self.stripe_size}B",
            ]
        else:
            return []

    def select_arg(self) -> list[str]:
        return self._select_arg(Subvolume.ORIG)

    def cache_select_arg(self) -> list[str]:
        return self._select_arg(Subvolume.CACHE)

    def _select_arg(self, subvolume: Subvolume) -> list[str]:
        pv_tag = self.pv_tag if subvolume is Subvolume.ORIG else self.cache_pv_tag
        if pv_tag:
            return [f"@{pv_tag}"]
        else:
            return []

    def type_arg(self) -> list[str]:
        if self.layout:
            return ["--type", ",".join(self.layout)]
        return []


@dataclass(frozen=True)
class VG:
    name: str
    pvs: dict[str | None, set[PV]] = field(default_factory=dict)
    lvs: dict[str, LV] = field(default_factory=dict)

    def get_lv_device(self, lv_name: str) -> Path:
        return Path("/dev") / self.name / lv_name


@dataclass
class LVMSetup:
    pvs: set[PV] = field(default_factory=set)
    vgs: dict[str, VG] = field(default_factory=dict)
    missing_pvs: set[PV] = field(default_factory=set)

    def is_empty(self) -> bool:
        return not self.pvs and not self.vgs

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> Self:
        check_type("lvm key", config, dict)
        entities = cls()
        for pv in config.get("pvs", []):
            check_type("lvm pv entry", pv, str)
            entities.pvs.add(
                PV(
                    device=Path(pv),
                )
            )
        for name, settings in config.get("vgs", {}).items():
            check_type(f"lvm vg {name} entry", settings, dict)

            lvs = settings.get("lvs", {})
            check_type(f"lvm vg {name} lvs entry", lvs, dict)

            lvs_entities = {}
            for lv_name, lv_settings in lvs.items():
                check_type(f"lvm vg {name} lv {lv_name} entry", lv_settings, dict)

                def handle_size(size_entry):
                    if size_entry == "rest":
                        return None
                    else:
                        return parse_size(size_entry, binary=True)

                size = handle_size(lv_settings["size"])
                cache_size = lv_settings.get("cache_size")
                cache_size = handle_size(cache_size) if cache_size is not None else None

                layout = {lv_settings.get("layout")}

                lvs_entities[lv_name] = LV(
                    name=lv_name,
                    layout=layout,
                    size=size,
                    stripes=lv_settings.get("stripes", 1),
                    stripe_size=parse_size(lv_settings.get("stripe_size", "0B")),
                    pv_tag=lv_settings.get("pv_tag"),
                    cache_pv_tag=lv_settings.get("cache_pv_tag"),
                    cache_size=lv_settings.get("cache_size"),
                    cache_mode=lv_settings.get("cache_mode"),
                )

            pvs = {}
            for tag, pvs_entry in settings.get("pvs", {}).items():
                check_type(f"vg {name} pvs entry", pvs_entry, list)
                pvs[tag] = {PV(device=Path(pv)) for pv in pvs_entry}

            entities.vgs[name] = VG(
                name=name,
                pvs=pvs,
                lvs=lvs_entities,
            )
        return entities

    @classmethod
    def from_system(cls, host: str) -> Self:
        # check if lvm2 is installed, return empty LVM entities otherwise
        if not exists_cmd("pvcreate", host=host, sudo=True):
            return cls()

        entities: Self = cls()

        lv_data = json.loads(
            run_cmd(
                [
                    "lvs",
                    "--all",
                    "--units",
                    "b",
                    "--options",
                    "lv_name,vg_name,lv_layout,lv_size,stripes,stripe_size,origin,cache_mode,pool_lv",
                    "--reportformat",
                    "json",
                ],
                host=host,
                sudo=True,
            ).stdout
        )["report"][0]["lv"]

        vg_devices_raw = json.loads(
            run_cmd(
                ["vgs", "--options", "vg_name,devices", "--reportformat", "json"],
                host=host,
                sudo=True,
            ).stdout
        )["report"][0]["vg"]
        vg_devices = defaultdict(list)
        for entry in vg_devices_raw:
            vg_devices[entry["vg_name"]].append(entry["devices"])

        vg_data = json.loads(
            run_cmd(
                ["vgs", "--options", "vg_name", "--reportformat", "json"],
                host=host,
                sudo=True,
            ).stdout
        )["report"][0]["vg"]

        pv_data = json.loads(
            run_cmd(
                [
                    "pvs",
                    "--options",
                    "pv_name,vg_name,lv_name,pv_tags",
                    "--reportformat",
                    "json",
                ],
                host=host,
                sudo=True,
            ).stdout
        )["report"][0]["pv"]

        for entry in vg_data:
            vg_name = entry["vg_name"]
            entities.vgs[vg_name] = VG(name=vg_name)

        for vg_name, device_reports in vg_devices.items():
            entities.missing_pvs.update(get_missing_pvs(device_reports))

        pv_tag_registry = {}
        for entry in pv_data:
            pv_device = entry["pv_name"]

            pv_tags = entry["pv_tags"].split(",")
            if len(pv_tags) > 1:
                raise UserError(
                    "Unsupported number of tags associated with "
                    f"PV {pv_device}: {pv_tags}. Only 1 or 0 allowed."
                )
            elif len(pv_tags) == 1:
                pv_tag = pv_tags[0]
            else:
                pv_tag = None

            pv_obj = PV(device=Path(pv_device))
            entities.pvs.add(pv_obj)
            vg_name = entry["vg_name"]
            if vg_name and vg_name in entities.vgs:
                # in case the PV is assigned to a vg
                vg_pvs = entities.vgs[vg_name].pvs
                if pv_tag not in vg_pvs:
                    vg_pvs[pv_tag] = set()
                vg_pvs[pv_tag].add(pv_obj)

            lv_name, is_internal = parse_lv_name(entry["lv_name"])
            if lv_name:
                pv_tag_registry[lv_name] = pv_tags[0]

        print(pv_tag_registry)

        internal_entries = defaultdict(dict)
        for entry in sorted(lv_data, key=lambda entry: entry["origin"]):
            print(entry)
            vg_name = entry["vg_name"]
            vg = entities.vgs[vg_name]
            lv_name, is_internal = parse_lv_name(entry["lv_name"])
            if is_internal:
                internal_entries[vg_name][lv_name] = entry
                continue

            origin, _ = parse_lv_name(entry["origin"])
            cache_mode = None
            if not origin:
                origin = None
            else:
                cache_mode = entry["cache_mode"]

            lv_orig_name = _subvolume_name(lv_name, Subvolume.ORIG)
            lv_cache_name = _subvolume_name(lv_name, Subvolume.CACHE)
            cache_entry = internal_entries.get(vg_name, {}).get(lv_cache_name, {})
            cache_size = cache_entry.get("lv_size")
            if cache_size is not None:
                cache_size = parse_size(cache_size)

            lv = LV(
                name=lv_name,
                layout=set(entry["lv_layout"].split(",")),
                size=parse_size(entry["lv_size"], binary=True),
                stripes=entry["stripes"],
                stripe_size=parse_size(entry["stripe_size"]),
                pv_tag=pv_tag_registry.get(lv_orig_name),
                cache_pv_tag=pv_tag_registry.get(lv_cache_name),
                cache_size=cache_size,
                cache_mode=cache_mode,
            )

            vg.lvs[lv_name] = lv
        return entities


def get_missing_pvs(device_reports: Iterable[str]) -> Iterable[PV]:
    for device_report in device_reports:
        m = VGS_DEVICE_REPORT_RE.match(device_report)
        assert m is not None, f"Invalid device report: {device_report}"
        info = m.group("info")
        if info == "missing":
            device = m.group("device")
            assert device is not None
            yield PV(device=Path(device))
        else:
            try:
                int(info)
            except ValueError:
                raise ValueError(f"Unexpected vgs device info: {info}")
